/* UADE - Unix Amiga Delitracker Emulator
 * Copyright 2000-2002, Heikki Orsila <heikki.orsila@ee.tut.fi>
 * See http://www.ee.tut.fi/~heikki/uade.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <dirent.h>

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "events.h"
#include "uae.h"
#include "memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "debug.h"

#include "uade.h"
#include "uade-os-unix.h"

#include "decrunch/decrunch.h"
#include "../plugindir/amifilemagic.h"

#include <netinet/in.h>

int uade_calc_reloc_size(unsigned int *src);

static FILE *uade_loadfile(char *fullname);
static void uade_putlong(int addr,int val);
static void uade_siginthandler(int);
static void uade_signalhandler(int);
static void uade_getmessage(void);
static void uade_waitformessage(int msg);
static void uade_setsubsongfunction(int);
static void uade_setntsc(int);
static void uade_setsongend(int songendpossible);
static void uade_sendamigamessage(int);

static void uade_enque_song(const char *filename);
static void check_name_extension(char *filename, char *playername);

static void uade_print_help(int problemcode);


uade_msgstruct* uade_struct;

char uade_playername[256];       // filename of eagleplayer
char uade_modulename[256];       // filename of song
char uade_scorename[256];        // filename of score file
char uade_mapname[256];          // filename of sharemem file

char score_playername[256];       // playername that eagleplayer reported (from emulation)
char score_modulename[256];       // modulename that eagleplayer reported
char score_formatname[256];	 // formatname that eagleplayer reported

int uade_setsubsongboolean = 0;
int uade_subsong = 0;
int uade_execdebugboolean = 0;
static int uade_dmawait = 0;
static int voltestboolean = 0; /* enabled if value is 0x12345678 */
int uade_force_by_default = 0;

static int playerforced = 0;

/* Is uade commanded by external application? */
int uade_slavemode = 0;
/* for cmdline tool streaming to another application */
int uade_usingoutpipe = 0;
/* This should be non-zero if native sound output shouldn't be used.
   This is set to non-zero in slavemode and outpipemode              */
int uade_no_native_sound = 0;

int uade_use_ntsc = 0;

static int uade_songendpossible = 1;

static int uade_repeat = 0;

int uade_reboot;

int uade_messages;       // number of messages in msgtable
static int msgtable[20]; // contains signals sent from master

static struct sigaction sigact;

int uade_mixpar = 0;

struct sn_t {
  struct sn_t *next;
  char filename[256];
};
typedef struct sn_t songnode_t;
songnode_t *SongList = 0;

int uade_max_subsong, uade_min_subsong, uade_cur_subsong;

/* contains uade's command line name */
static char *uadecmdlinename = 0;


void uade_option(int argc, char** argv) {
  int i;

  uadecmdlinename = strdup(argv[0]);

  uade_mapname[0] = 0;
  uade_playername[0] = 0;

  uade_get_path(uade_scorename, UADE_PATH_SCORE);

  for (i = 1; i < argc;) {
    // check for playername parameter
    if (strcasecmp(argv[i], "-P") == 0) {
      if((i+1) < argc) {
	strcpy(uade_playername, argv[i+1]);
	playerforced = 1;
	fprintf(stderr,"-P parameter: %s\n", uade_playername);
	i+=2;
      } else {
	fprintf(stderr, "Player name missing for -P parameter\n\n");
	uade_print_help(1);
	exit(-1);
      }

      // check for modulename parameter
    } else if (strcasecmp(argv[i], "-M") == 0) {
      if((i+1) < argc) {
	fprintf(stderr, "-M parameter: %s\n", argv[i+1]);
	uade_enque_song(argv[i+1]);
	i+=2;
      } else {
	fprintf(stderr, "Song name missing for -M parameter\n\n");
	uade_print_help(1);
	exit(-1);
      }

      // check for scorename parameter
    } else if(strcasecmp(argv[i], "-S") == 0) {
      strcpy(uade_scorename, argv[i+1]);
      fprintf(stderr, "-S parameter: %s\n", uade_scorename);
      i+=2;

      // check if this prog is requested for slave
      // that let's external programs use this program as
      // a player
    } else if (strcmp(argv[i], "-slave") == 0) {
      strcpy(uade_mapname, argv[i+1]);
      uade_slavemode = 1;
      uade_no_native_sound = 1;
      i+=2;

    } else if (strcmp(argv[i], "-outpipe") == 0) {
      fprintf(stderr, "-outpipe parameter: %s\n", argv[i+1]);
      uade_usingoutpipe = 1;
      uade_no_native_sound = 1;
      uade_init_outpipe(argv[i+1]);
      i+=2;

    } else if (strcmp(argv[i], "-force") == 0) {
      fprintf(stderr, "-force parameter\n");
      uade_force_by_default = 1;
      i++;

    } else if (strcmp(argv[i], "-execdebug") == 0) {
      fprintf(stderr, "-execdebug parameter\n");
      uade_execdebugboolean = 1;
      i++;

    } else if (strcmp(argv[i], "-voltest") == 0) {
      fprintf(stderr, "-voltest parameter\n");
      voltestboolean = 0x12345678;
      i++;

    } else if (strcmp(argv[i], "-dmawait") == 0) {
      uade_dmawait = atoi(argv[i+1]);
      fprintf(stderr, "setting dmawait = %d rasterlines\n", uade_dmawait);
      uade_dmawait += 0x12340000;
      i += 2;

    } else if (strcmp(argv[i], "-ntsc") == 0) {
      fprintf(stderr, "-ntsc parameter (default is PAL)\n");
      uade_use_ntsc = 1;
      i++;

    } else if (strcmp(argv[i], "-noend") == 0) {
      fprintf(stderr, "-noend parameter (player can't force song to end\n");
      uade_songendpossible = 0;
      i++;

    } else if (strcmp(argv[i], "-sub") == 0) {
      uade_subsong = atoi(argv[i+1]);
      if((*argv[i+1] != '+') && (*argv[i+1] != '-')) { 
	uade_setsubsongboolean = 1;
	fprintf(stderr, "-sub parameter parsed, setting subsong %d\n", uade_subsong);
      } else {
	uade_setsubsongboolean = 2;
	fprintf(stderr, "-sub parameter parsed, setting subsong %d+min_subsong\n", uade_subsong);
      }
      i+=2;

      /* -int is obsolete */
    } else if (strcmp(argv[i], "-int") == 0) {
      fprintf(stderr, "-int is an obsolete parameter!\n\n");
      i++;

    } else if (strcmp(argv[i], "--verbose") == 0) {
      i++;

    } else if (strcmp(argv[i], "--version") == 0) {
      fprintf(stdout,"UADE %s\n", uade_version());
      exit(0);

    } else if (strcmp(argv[i], "-varpid") == 0) {
      /* ADDED BY XIGH */
      uade_create_var_pid();
      i++;

    } else if(strcmp(argv[i], "-pan")==0) {
      fprintf(stderr, "-pan parameter parsed (%s)\n", argv[i+1]);
      uade_mixpar = (int) (256.0f/2.0f*atof(argv[i+1]));
      i+=2;

    } else if(strcmp(argv[i], "-config")==0) {
      i++;

    } else if(strcmp(argv[i], "-repeat")==0) {
      uade_repeat = 1;
      i++;

    } else if(strcmp(argv[i], "--help")==0 || strcmp(argv[i], "-h")==0) {
      uade_print_help(0);
      exit(0);

    } else {

      uade_enque_song(argv[i]);

      i++;
    }
  }

  if(!uade_slavemode) {
    // if songlist is empty
    if(!SongList) {
      if(uade_playername[0] != 0) {
	uade_enque_song(uade_playername);
      } else {
	uade_print_help(2);
	exit(-1);
      }
    }
    /* create CTRL-C break for cmdline tool */
    uade_create_signalhandler(uade_siginthandler, 2);
  }

  if(uade_slavemode) {

    uade_struct = (uade_msgstruct*) uade_mmap_file(uade_mapname);
    if(!uade_struct) {
      fprintf(stderr, "uade.c/uade: couldn't mmap file\n");
      exit(EXIT_FAILURE);
    }

    uade_struct->minsubsong = 0;
    uade_struct->maxsubsong = 0;
    uade_reboot = 1;

    /* clear messages in msgtable */
    uade_messages = 0;
    for(i=0; i<20; i++) msgtable[i] = 0;

    /* setup a signal handler for communication with xmms plugin */
    uade_create_signalhandler(uade_signalhandler, 1);

    /* leave a watchdog that checks if XMMS plugin dies (polls plugin
       task with a kill(pluginpid,0) signal every 0.5 seconds). If XMMS
       plugin dies, watchdog kills UAE and itself */
    uade_create_watchdog(uade_struct, uade_mapname);
  }
}


static void uade_print_help(int problemcode) {
  switch(problemcode) {
  case 0:
    /* just for printing help */
    break;
  case 1:
    fprintf(stderr, "Illegal parameters for UADE\n\n");
    break;
  case 2:
    fprintf(stderr, "No songs given as parameters for UADE\n\n");
    break;
  default:
    fprintf(stderr, "Unkown error\n");
    break;
  }
  fprintf(stderr, "UADE usage:\n");
  fprintf(stderr, " %s [OPTIONS] [FILE1] [FILE2] ...\n\n", uadecmdlinename);

  fprintf(stderr, "options:\n");
  fprintf(stderr, " -M songname\tSets songname to be played\n");
  fprintf(stderr, " -P playername\tSets eagleplayer to be used (you can use this for custom songs)\n");
  fprintf(stderr, " -sub number\tSets subsong number to be played\n");
  fprintf(stderr, " -force\t\tForces eagleplayer to play a given song if it's not recognized\n");
  fprintf(stderr, " -repeat\tLoop song list to be played from beginning after the end\n");
  fprintf(stderr, " -pan\t\tSets panning value in the range of [0.00, 1.00] (mix channels)\n");
  fprintf(stderr, " -noend\t\tThe first song will be played indefinetely (probably)\n");
  fprintf(stderr, " -ntsc\t\tForce Amiga into NTSC mode (may affect timing)\n");
  fprintf(stderr, " -outpipe x\tDump all sound output to file descriptor x (x=0 => stdout)\n"); 
  fprintf(stderr, " -S scorename\tForces scorename file to be used as 'sound core' (experts only)\n");
}


static void uade_enque_song(const char *fname) {
  songnode_t *s = SongList;
  songnode_t *new;
  if(uade_if_normal_file(fname)) {
    new = malloc(sizeof(songnode_t));
    new->next = 0;
    strcpy(new->filename, fname);
    if(s) {
      while(s->next) s = s->next;
      s->next = new;
    } else {
      SongList = new;
      }
  } else {
    fprintf(stderr,"uade: %s is not regular file\n", fname);
  }
}


static void check_name_extension(char *filename, char *playername) {
  FILE *formatsfile;
  FILE* songfile;
  int filesize = 5122;
  unsigned char buf[5122]="";
  int status;
  char formatsfilename[256];
  char playerdir[256];
  int x;
  char a[256],b[256],pre[256],post[256];
  char *c,*filename2;

  *playername = 0;

  if(!uade_get_path(formatsfilename,UADE_PATH_FORMATSFILE))
    return;
  formatsfile = fopen(formatsfilename, "rb");

  /* get player directory */
  uade_get_path(playerdir, UADE_PATH_PLAYERDIR);

  pre[0]='\0';
  post[0]='\0';

  filename2 = uade_get_basename(filename);
  c = strchr(filename2, (int) '.');
  if(c == 0) {
    goto closeffile;
  }
  strncpy(pre, filename2, c-filename2);
  pre[c-filename2] = '\0';              // make so that pre is null-terminated
  c = strrchr(filename2, (int) '.');
  strcpy(post, c+1);                    // post is null-terminated

  songfile = fopen (filename, "rb");
  if (!songfile) {
    fprintf (stderr, "can not open file %s\n", filename);
    return;
  }
  if ((status = decrunch ( &songfile,filename ) < 0 )) {
    fprintf (stderr, "decrunching error...\n");
    fclose (songfile);
    return;
  }
  fread  (buf, 1, 5122, songfile);
  fclose (songfile);
  filemagic (buf, pre, post);

  for(;;) {
    x = fscanf(formatsfile,"%s",a);
    if(x==0 || x==EOF) break;

    // check if this is a comment line (this skips comment lines)
    if(a[0]=='#') {
      while(a[0]!='\n') {
	x=fscanf(formatsfile,"%c",&a[0]);
	if(x==0 || x==EOF) break;
      }
      continue;
    }

    if(strcasecmp(a,"formats")==0) {
      for(;;) {
	x=fscanf(formatsfile,"%s",a);
	if(x==0 || x==EOF) break;
	if(strcasecmp("endformats",a)==0) break;
	
	// check if this is a comment line (this skips comment lines)
	if(a[0]=='#') {
	  while(a[0]!='\n') {
	    x=fscanf(formatsfile,"%c",&a[0]);
	    if(x==0 || x==EOF) break;
	  }
	  continue;
	}
	
	x=fscanf(formatsfile,"%s",b);
	if(x==0 || x==EOF) break;

	if(strcasecmp(pre,a)==0 || strcasecmp(post,a)==0) {
	  if(strcmp(b,"custom")==0) {
	    strcpy(playername, filename);
	    filename[0] = 0;
	  } else {
	    /* Find out player directory path */
	    strcpy(playername, playerdir);
	    strcat(playername, b);
	  }
	  goto closeffile;
	}
      }
      break;
    }
  }
 closeffile:
  fclose(formatsfile);
}


/* catch CTRL-C and cause REBOOT */
UADE_SIGNALHANDLER(uade_siginthandler) {
  int ltime;
  signum = 0;
  ltime = uade_timer_poll();
  if(uade_reboot || (ltime>0 && ltime<100)) {
    fprintf(stderr,"uade: exit forced with ctrl-c\n");
    exit(0);
  }
  uade_reboot = 1;
}


UADE_SIGNALHANDLER(uade_signalhandler) {
  int msg,msgs;
  char a = 0;
  int i;

  msg = uade_struct->touaemsgtype;
  uade_struct->touaemsgtype = 0;

  switch(msg) {
  case UADE_REBOOT:
    if(uade_reboot==1) return;
    uade_reboot = 1;
    uade_clearsoundbuffer();
    //fprintf(stderr,"Received UADE_Reboot!\n");
    break;
  case UADE_SETSUBSONG:
    //fprintf(stderr,"Received UADE_SETSUBSONG (%d)\n",uade_struct->subsong);
    uade_setsubsongfunction(uade_struct->subsong);
    uade_struct->setsubsongboolean = 0;
    return;

  case UADE_PLAYERNAME:
  case UADE_MODULENAME:
  case UADE_SCORENAME:
    break;
  case UADE_NTSC:
    uade_setntsc(uade_struct->ntscbit);
    uade_sendamigamessage(UADE_NTSC);
    //fprintf(stderr,"Received msg telling NTSC = %d\n",uade_struct->ntscbit);
    return;
  case UADE_SONGEND:
    uade_songendpossible = uade_struct->songendpossible;
    uade_setsongend(uade_songendpossible);
    //fprintf(stderr,"Received msg telling songendpossible = %d\n",uade_songendpossible);
    return;
  default:
    fprintf(stderr,"uade_getmessage: Msg was not recognized (%d)\n",msg);
    break;
  }

  msgtable[msg]=1;

  // count messages in msgtable
  msgs=0;
  for(i=0;i<20;i++) msgs+=(msgtable[i]==0)?0:1;
  uade_messages=msgs;
  //fprintf(stderr,"uae: getmessage: uade_messages=%d\n",uade_struct->messages);
}

void uade_waitformessage(int msg) {
  int sl = 0;
  //  fprintf(stderr,"uade: uade_waitformessage: %d\n",msg);
  while(1) {
    if(msgtable[msg] != 0) {
      uade_handlemessage(msg);
      break;
    }
    uade_usleep(1000);
    sl += 1000;
  }
  fprintf(stderr,"uade/uade_waitformessage: slept %d\n",sl/1000);
}


void uade_handlemessage(int msg) {
  int i,msgs;
  msgtable[msg] = 0;
  msgs = 0;
  for(i=0;i<20;i++)
    msgs += (msgtable[i]==0) ? 0:1;
  uade_messages=msgs;
}


void uade_prerun(void) {
  /* don't load anything under 0x1000 (exec at $1000) */
  int modnameaddr = 0x00400;
  int scoreaddr   = 0x01000;
  int userstack   = 0x08500;
  int superstack  = 0x08f00;
  int playeraddr  = 0x09000;
  int relocaddr   = 0x40000;   /* default, recalculate this */
  int modaddr     = 0x80000;   /* default, recalculate this */
  int len;
  FILE* file;
  uae_u8* memp;
  int bytesread;
  // decrunch variables
  FILE* file_cr;
  char *modfile;
  int status;

  /* IMPORTANT:
     It seems that certain players don't work totally reliably if memory contains
     trash from previous songs. To be certain that each song is played from the
     same initial state of emulator we clear the memory from 0x400 to 0x100000
     each time a new song is played */
  memp = get_real_address(0x400);
  memset(memp, 0, 0x100000-0x400);

  if(uade_slavemode) {

    /* RACE CONDITION BEGINS */
    /* uade executable is ready to receive commands */
    uade_struct->uade_inited_boolean = 1;
    /* sleep() is interrupted by a SIGHUP from XMMS plugin */
    while(uade_struct->loadnewsongboolean == 0) uade_sleep(2);
    /* RACE CONDITION ENDS */

    strcpy(uade_playername, uade_struct->playername);
    strcpy(uade_modulename, uade_struct->modulename);
    strcpy(uade_scorename, uade_struct->scorename);

    uade_struct->loadnewsongboolean = 0;

    uade_force_by_default = uade_struct->force_by_default;
    if(uade_struct->setsubsongboolean) {
      uade_struct->setsubsongboolean = 0;
      uade_setsubsongboolean = 1;
      uade_subsong = uade_struct->subsong;
    }
    uade_struct->minsubsong = uade_struct->maxsubsong = 0;
    uade_songendpossible = uade_struct->songendpossible;

  } else {

  takenextsong:

    /* if no more songs in the play queue, quit cmdline tool */
    if(!SongList) {
      fprintf(stderr,"uade: no more songs to play\n");
      exit(EXIT_SUCCESS);

    } else {
      char playername[256];
      char tempname[256];
      strcpy(uade_modulename,SongList->filename);
      strcpy(tempname,SongList->filename);

      /* uade can be forced to use 'uade_playername' initialized with
	 -P switch for all played songs. */
      if(!playerforced) {
	check_name_extension(uade_modulename, playername);
	/* playername is empty string if no playername was found */
	if(playername[0]!=0) {
	  strcpy(uade_playername, playername);
	} else {
	  fprintf(stderr,"uade: file %s: unknown format\n", SongList->filename);
	}
      }
      SongList = SongList->next;

      if(uade_repeat) {
	uade_enque_song(tempname);
      }
    }

    if(uade_playername[0]==0) {
      goto takenextsong;
    }

    uade_cur_subsong = uade_min_subsong = uade_max_subsong = 0;
  }

  /* set songendpossible to amiga memory*/
  uade_setsongend(uade_songendpossible);

  /* these are for debugging purposes */
  if(uade_execdebugboolean) {
    uade_putlong(0x18C, 0x12345678);
  } else {
    uade_putlong(0x18C, 0);
  }
  uade_putlong(0x190,voltestboolean);
  uade_putlong(0x194,uade_dmawait);

  /* open eagleplayer */
  file = fopen(uade_playername, "rb");
  if(file) {
    if ((status = decrunch (&file, uade_playername)) < 0) {
      fprintf (stderr, "decrunching error...");
      fclose(file);
      goto setupfailure;
    } else {
      memp = get_real_address(playeraddr);
      bytesread = 0;
      while(!feof(file)) {
	bytesread += fread(memp+bytesread, 1, 65536, file);
      }
      fprintf(stderr,"player '%s' (%d bytes)\n",uade_playername,bytesread);
      fclose(file);
      // set player executable address for relocator
      uade_putlong(0x108, playeraddr);

      len = uade_calc_reloc_size((unsigned int *) memp);
      if(len) {
	relocaddr  = ((playeraddr + bytesread) & 0x7FFFF000) + 0x2000;
	modaddr = ((relocaddr + len) & 0x7FFFF000) + 0x2000;
      } else {
	fprintf(stderr, "uae/uade.c: problem with reloc calculation\n");
	fprintf(stderr, "uae/uade.c: setting default addresses\n");
	// set address to relocate player to ...
      }
      // fprintf(stderr, "uae/uade.c: using relocaddr = 0x%x moduleaddr = 0x%x\n", relocaddr, modaddr);
      uade_putlong(0x10c, relocaddr);
    }
  } else {
    fprintf(stderr,"uade: couldn't load player %s\n", uade_playername);
  setupfailure:
    if(uade_slavemode) {
      uade_reboot = 0;
      uade_struct->songendboolean = 1;
      return;
    }
    exit(EXIT_FAILURE);
  }

  /* load module (not necessary) */
  uade_putlong(0x100, modaddr);     /* set modules address */
  uade_putlong(0x104, 0);           /* set module size to zero */
  uade_putlong(0x128, 0);           /* mod name address pointer */
  if(uade_modulename[0] != 0) {
    if((file = fopen(uade_modulename, "rb"))) {
      //fprintf (stderr, "testing if file is crunched...\n");
      if ((status = decrunch (&file, uade_modulename)) < 0) {
	fprintf (stderr, "decrunching error...");
	fclose(file);
	goto setupfailure;
      } else {
	/* load file*/
	memp = get_real_address(modaddr);
	bytesread = 0;
	while(!feof(file)) {
	  bytesread += fread(memp+bytesread,1,65536,file);
	}
	fprintf(stderr, "module '%s' (%d bytes)\n", uade_modulename,bytesread);
	fclose(file);
	// set module size
	uade_putlong(0x104, bytesread);
	memp = get_real_address(modnameaddr);
	strncpy(memp, uade_modulename, 255);
	uade_putlong(0x128, modnameaddr);
      }
    } else {
      goto setupfailure;
    }
  } else {
    memp = get_real_address(modnameaddr);
    strncpy(memp, uade_playername, 255);
    uade_putlong(0x128, modnameaddr);
  }

  /* load sound core (score) */
  if((file = fopen(uade_scorename, "rb"))) {
    memp = get_real_address(scoreaddr);
    bytesread = 0;
    while(!feof(file)) {
      bytesread += fread(memp+bytesread,1,65536,file);
    }
    fprintf(stderr,"score  '%s' (%d bytes)\n",uade_scorename,bytesread);
    fclose(file);
    // set Amiga running from begin of the soundcore
    m68k_setpc(scoreaddr);
    // override bit for sound format checking
    uade_putlong(0x118, uade_force_by_default);
    // setsubsong
    uade_putlong(0x11c, uade_setsubsongboolean);
    uade_putlong(0x120, uade_subsong);
    /* set ntscbit correctly */
    if(uade_slavemode!=0) {
      uade_setntsc(uade_struct->ntscbit);
    } else {
      uade_setntsc(uade_use_ntsc);
    }
    // pause bits (don't care!), for debugging purposes only
    uade_putlong(0x180, 0);
    uade_putlong(0x184, 0);
    // set user and supervisor stack pointers
    uade_putlong(0x110, userstack);
    uade_putlong(0x114, superstack);
    // input msg handler flag to zero
    uade_putlong(0x300, 0);

    if( (userstack - (scoreaddr + bytesread)) < 0x1000) {
      fprintf(stderr, "uae/uade.c: stack over run warning!\n");
    }

  } else {
    goto setupfailure;
  }

  uade_setsubsongboolean = 0;

  if(!playerforced) {
    uade_playername[0] = 0;
  }
  uade_modulename[0] = 0;
  score_playername[0] = 0;
  score_modulename[0] = 0;
  if(uade_slavemode) {
    uade_struct->score_playername[0]='\0';
    uade_struct->score_modulename[0]='\0';
  }

  // don't reboot after this routine has run unless SIGNAL has been
  // caught...
  uade_reboot = 0;
  if(uade_slavemode) {
    if(msgtable[UADE_REBOOT]) {
      int l;
      for(l=0;l<20;l++) {
	fprintf(stderr,"uade/uade.c: TRIED TO REBOOT 2 TIMES !!!\n");
	fprintf(stderr," If you can make this bug happen again please submit");
	fprintf(stderr," a bug report.\n");
      }
    }
  }
}

static void uade_putlong(int addr,int val) {
  uae_u8 *p;
  p=get_real_address(addr+0);
  *p = (uae_u8) (val>>24 & 0xff);
  p=get_real_address(addr+1);
  *p = (uae_u8) (val>>16 & 0xff);
  p=get_real_address(addr+2);
  *p = (uae_u8) (val>>8 & 0xff);
  p=get_real_address(addr+3);
  *p = (uae_u8) (val & 0xff);
}

int uade_getlong(int adr) {
  uae_u8* ptr;
  int x=0;
  ptr=get_real_address(adr);
  x+=(*ptr)<<24;
  ptr=get_real_address(adr+1);
  x+=(*ptr)<<16;
  ptr=get_real_address(adr+2);
  x+=(*ptr)<<8;
  ptr=get_real_address(adr+3);
  x+=(int) (*ptr);
  return x;
}

void uade_getamigamessage(void) {
  uae_u8* ptr;
  uae_u8* nameptr;
  int x;
  int mins,maxs,curs;
  FILE *file;
  int size;
  int status;

  // GET MESSAGE TYPE FROM VIRTUAL ADDR $200
  x = uade_getlong(0x200);

  switch(x) {
  case UADE_SONGEND:
    fprintf(stderr,"uae/uade.c: song end (player)\n");
    if(uade_slavemode) {
      uade_struct->songendboolean = 1;
    } else {
      if(uade_max_subsong) {
	if(uade_cur_subsong<uade_max_subsong) {
	  uade_cur_subsong++;
	  uade_setsubsongfunction(uade_cur_subsong);
	  break;
	}
      }
      uade_reboot = 1;
    }
    break;
  case UADE_SUBSINFO:
    mins=uade_getlong(0x204);
    maxs=uade_getlong(0x208);
    curs=uade_getlong(0x20C);
    fprintf(stderr,"uae/uade.c: subsong info: min subsong: %d max subsong: %d cur subsong: %d\n",mins,maxs,curs);
    if(uade_slavemode) {
      uade_struct->minsubsong = mins;
      uade_struct->maxsubsong = maxs;
      uade_struct->subsong = curs;
    } else {
      uade_min_subsong = mins;
      uade_max_subsong = maxs;
      uade_cur_subsong = curs;
    }
    break;
  case UADE_PLAYERNAME:
    ptr = get_real_address(0x204);
    strncpy(score_playername,ptr,25);
    fprintf(stderr,"uae/uade.c: playername: %s\n",score_playername);
    if(uade_slavemode) {
      strncpy(uade_struct->score_playername,score_playername,25);
    }
    break;
  case UADE_MODULENAME:
    ptr = get_real_address(0x204);
    strncpy(score_modulename,ptr,25);
    fprintf(stderr,"uae/uade.c: modulename: %s\n",score_modulename);
    if(uade_slavemode) {
      strncpy(uade_struct->score_modulename,score_modulename,250);
    }
    break;
  case UADE_FORMATNAME:
    ptr = get_real_address(0x204);
    strncpy(score_formatname,ptr,255);
    fprintf(stderr,"uae/uade.c: formatname: %s\n",score_formatname);
    if(uade_slavemode) {
      strncpy(uade_struct->score_formatname,score_formatname,255);
    }
    break;
  case UADE_GENERALMSG:
    fprintf(stderr,"uae/uade.c: general message: %s\n",get_real_address(0x204));
    break;
  case UADE_CHECKERROR:
    fprintf(stderr,"uae/uade.c: FAILURE: Module check FAILED\n");
    if(!uade_slavemode) uade_reboot = 1;
    break;
  case UADE_SCORECRASH:
    fprintf(stderr,"uae/uade.c: FAILURE: Soundcore CRASHED\n");
    if(uade_slavemode)
      uade_struct->songendboolean = 1;
    else
      uade_reboot = 1;
    break;
  case UADE_SCOREDEAD:
    fprintf(stderr,"uae/uade.c: FAILURE: Soundcore is in DEADEND\n");
    if(uade_slavemode) {
      uade_struct->songendboolean = 1;
    } else {
      uade_reboot = 1;
    }
    break;
  case UADE_LOADFILE:
    /* load a file named at 0x204 (name pointer) to address pointed by
       0x208 and insert the length to 0x20C */
    nameptr = get_real_address(uade_getlong(0x204));
    if((file = uade_open_amiga_file(nameptr))) {
      if ((status = decrunch (&file, nameptr)) < 0) {
	fprintf (stderr, "UADE_LOADFILE: decrunching error\n");
	fclose(file);
	break;
      } else {
	ptr = get_real_address(uade_getlong(0x208));
	size = 0;
	while(!feof(file)) {
	  size += fread(ptr+size, 1, 65536, file);
	}
	fclose(file);
	uade_putlong(0x20C,size);
	fprintf(stderr, "UADE_LOADFILE: %s: ptr = 0x%x size = 0x%x\n", nameptr, uade_getlong(0x208), size);
      }
    }
    break;

  case UADE_READ:
    nameptr = get_real_address(uade_getlong(0x204));
    fprintf(stderr,"UADE_READ: '%s' dst = 0x%x off = 0x%x len = 0x%x\n", nameptr, uade_getlong(0x208), uade_getlong(0x20C), uade_getlong(0x210));
    if((file = uade_open_amiga_file(nameptr))) {
      fseek(file, uade_getlong(0x20C), SEEK_SET);
      size = fread(get_real_address(uade_getlong(0x208)), 1, uade_getlong(0x210), file);
      fclose(file); file = 0;
      if(size != uade_getlong(0x210))
	fprintf(stderr, "UADE_READ: 0x%x bytes read\n", size);
      uade_putlong(0x214, size);
    } else {
      fprintf(stderr, "UADE_READ: error when reading '%s'\n", nameptr);
      uade_putlong(0x214, 0);
    }
    break;

  case UADE_FILESIZE:
    nameptr = get_real_address(uade_getlong(0x204));
    if((file = uade_open_amiga_file(nameptr))) {
      fseek(file, 0, SEEK_END);
      size = ftell(file);
      fclose(file); file = 0;
      fprintf(stderr, "UADE_SIZE: 0x%x '%s'\n", size, nameptr);
      uade_putlong(0x208, size);
      uade_putlong(0x20C, -1);
    } else {
      fprintf(stderr,"UADE_SIZE: failure\n");
      uade_putlong(0x208, 0);
      uade_putlong(0x20C, 0);
    }
    break;

  default:
    fprintf(stderr,"uae/uade.c: UNKNOWN MESSAGE from Amiga (trap #5): ADDR $200: $%x\n",x);
    break;
  }
}


static void uade_setsubsongfunction(int subs) {
  uade_putlong(0x120,subs);
  uade_sendamigamessage(UADE_SETSUBSONG);
  uade_clearsoundbuffer();
}

static void uade_setntsc(int usentsc) {
  uade_putlong(0x124,usentsc);
}

static void uade_setsongend(int songendpossible) {
  uade_putlong(0x12C,songendpossible);
}

static void uade_sendamigamessage(int msgtype) {
  uade_putlong(0x300,msgtype);
}

void uade_clearsoundbuffer(void) {
  if(uade_slavemode) {
    memset(uade_struct->soundbuffer, 0, sizeof(uade_struct->soundbuffer));
  }
}


int uade_calc_reloc_size(unsigned int *src) {
  unsigned int offset;
  int i,nhunks;
  if(sizeof(unsigned int) != 4) {
    fprintf(stderr, "uade: reloc size problem: unsigned int != 4\n");
    fprintf(stderr, "uade: make reloc size calculator more portable\n");
    return 0;
  }
  if(ntohl(*src) != 0x03f3) return 0;
  src++;
  if(ntohl(*src) != 0) return 0;
  src++;
  nhunks = ntohl(*src); src++; /* take number of hunks */
  src+=2;                      /* skip first & last hunk indexes */
  offset = 0;
  for(i = 0; i < nhunks; i++) {
    offset += 4 * (ntohl(*src) & 0x3FFFFFFF); src++;
  }
  return ((int) offset);
}
